/*
 *  Copyright (C) 2007-11 Luca Deri <deri@ntop.org>
 *
 *  		       http://www.ntop.org/
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_PF_RING

/* ****************************************************** */

inline void processPfringPktHdr(struct pfring_pkthdr *hdr, char *packet, long thread_id) {
  decodePacket(thread_id, (struct pcap_pkthdr*)hdr, packet,
	       0 /* sampledPacket */,
	       (hdr->extended_hdr.parsed_pkt.pkt_detail.flow.samplingPopulation == 0) ?
	       1 /* numPkts */ : hdr->extended_hdr.parsed_pkt.pkt_detail.flow.samplingPopulation,
	       (readOnlyGlobals.inputInterfaceIndex == DEFAULT_INPUT_INTERFACE_INDEX) ? hdr->extended_hdr.parsed_pkt.pkt_detail.flow.in_iface : NO_INTERFACE_INDEX,
	       (readOnlyGlobals.outputInterfaceIndex == DEFAULT_INPUT_INTERFACE_INDEX) ? hdr->extended_hdr.parsed_pkt.pkt_detail.flow.out_iface : NO_INTERFACE_INDEX,
	       0 /* flow_sender_ip */);
}

/* ****************************************************** */

struct n2disk_metadata_header {
  u_int16_t version;
  u_int16_t metadata_len;
};

struct n2disk_metadata {
  u_int32_t pkt_offset;
  struct pfring_pkthdr metadata;
};

static void* readMetadataPkts(void) {
  struct n2disk_metadata_header metadata_hdr;
  struct n2disk_metadata entry;
  int n;

  n = fread((void*)&metadata_hdr, 1, sizeof(metadata_hdr), readOnlyGlobals.metadata_fd);
  if(n != sizeof(metadata_hdr)) {
    traceEvent(TRACE_WARNING, "Metadatda file is too short");
    return(NULL);
  }

  while(fread((void*)&entry, 1, sizeof(entry), readOnlyGlobals.metadata_fd) == sizeof(entry)) {
    processPfringPktHdr(&entry.metadata, NULL, 0 /* fake thread_id */);
  }

  return(NULL);
}

/* ****************************************************** */

void printPfRingStats(void) {
  int i;

  for(i=0; i<readOnlyGlobals.numProcessThreads; i++) {
    if(readWriteGlobals->ring[i] != NULL) {
      pfring_stat stats;
      
      if(pfring_stats(readWriteGlobals->ring[i], &stats) >= 0)
	traceEvent(TRACE_NORMAL, "Packet stats (PF_RING) [%d]: %u/%u pkts rcvd/dropped",
		   i, stats.recv, stats.drop);
    }

    if(readOnlyGlobals.clusterId == 0)
      break;      
  }
}

/* ****************************************************** */

void* fetchPfRingPackets(void* notUsed) {
  unsigned long thread_id = (unsigned long)notUsed;
  struct pfring_pkthdr hdr;
  u_char *packet;
  int rc, use_full_packet = 0;
  struct pcap_pkthdr h;

  if(readOnlyGlobals.metadata_fd) return(readMetadataPkts());

#if 0
#ifdef HAVE_PTHREAD_SET_AFFINITY
  if((getNumCores() > 1) && (readOnlyGlobals.numProcessThreads > 1)) {
    /* Bind this thread to a specific core */
    cpu_set_t cpuset;
    int rc, core_id = thread_id % readOnlyGlobals.numProcessThreads;

    CPU_ZERO(&cpuset);
    CPU_SET(core_id, &cpuset);

    if((rc = pthread_setaffinity_np(pthread_self(), sizeof(cpu_set_t), &cpuset)) != 0)
      traceEvent(TRACE_ERROR, "Error while binding to core %ld: errno=%i\n",
		 core_id, rc);
    else {
      traceEvent(TRACE_INFO, "Bound thread to core %lu/%u\n", core_id, getNumCores());
    }
  }
#endif
#endif

  packet = (u_char*)malloc(readOnlyGlobals.snaplen+1);

  if(packet == NULL) {
    traceEvent(TRACE_WARNING, "Not enough memory!");
    return(NULL);
  }

  if(readOnlyGlobals.pktSampleRate > 1)
    rc = pfring_set_sampling_rate(readWriteGlobals->ring[thread_id], readOnlyGlobals.pktSampleRate);

  while(!readWriteGlobals->shutdownInProgress) {
    rc = pfring_recv(readWriteGlobals->ring[thread_id], 
		     &packet, readOnlyGlobals.snaplen,
		     &hdr, 1 /* wait_for_incoming_packet */);

    if(rc > 0)
      processPfringPktHdr(&hdr, packet, thread_id);
  }

  free(packet);

  return(NULL);
}

/* ********************************************* */

pfring* open_ring(char *dev, u_char *open_device, u_short thread_id) {
  pfring* the_ring = NULL;

  /*
    We disable promiscuous mode when using NFlite as we will capture
    just packets sent to use and not those that belong to other
    host (thus that are not interesting for us)
  */
  if(readOnlyGlobals.nfLitePluginEnabled)
    readOnlyGlobals.promisc_mode = 0, readOnlyGlobals.snaplen = 256;

  if((the_ring = pfring_open(dev,
			     readOnlyGlobals.promisc_mode /* promiscuous */,
			     readOnlyGlobals.snaplen,
			     ((readOnlyGlobals.clusterId == 0) && (readOnlyGlobals.numProcessThreads > 1))
			     ? 1 /* reentrant */ : 0)) != NULL) {
    u_int32_t version;
    int rc;

    rc = pfring_version(the_ring, &version);

    if((rc == -1) || (version < 0x030502)) {
      traceEvent(TRACE_WARNING,
		 "nProbe requires PF_RING v.3.9.3 or above (you have v.%d.%d.%d)",
		 (version & 0xFFFF0000) >> 16,
		 (version & 0x0000FF00) >> 8,
		 version & 0x000000FF);
      pfring_close(the_ring);
      the_ring = NULL;
    } else {
      if(thread_id == 0)
	traceEvent(TRACE_INFO, "Successfully open PF_RING v.%d.%d.%d on device %s [snaplen=%u]\n",
		   (version & 0xFFFF0000) >> 16, (version & 0x0000FF00) >> 8,
		   (version & 0x000000FF),
		   readOnlyGlobals.captureDev, readOnlyGlobals.snaplen);
      *open_device = 0;
      readOnlyGlobals.datalink = DLT_EN10MB;
      pfring_set_application_name(the_ring, "nProbe");

      if((readOnlyGlobals.numProcessThreads > 1)
	 && (readOnlyGlobals.clusterId > 0))
	pfring_set_cluster(the_ring, readOnlyGlobals.clusterId, cluster_per_flow);

      if(thread_id == 0) {
	traceEvent(TRACE_NORMAL, "Using PF_RING in-kernel accelerated packet parsing");

	if(readOnlyGlobals.clusterId > 0)
	  traceEvent(TRACE_NORMAL, "PF_RING socket bound to cluster %d", readOnlyGlobals.clusterId);
      }

      if(readOnlyGlobals.nfLitePluginEnabled) {
	filtering_rule rule;

	pfring_toggle_filtering_policy(the_ring, 0); /* Default to drop */

	memset(&rule, 0, sizeof(rule));

	rule.rule_id = 1;
	rule.rule_action = execute_action_and_stop_rule_evaluation;
	rule.plugin_action.plugin_id = 13 /* NFLITE_PLUGIN_ID */;
	rule.core_fields.proto       = 17; /* UDP */
	rule.core_fields.port_low    = readOnlyGlobals.nfLitePluginLowPort;
	rule.core_fields.port_high   = readOnlyGlobals.nfLitePluginLowPort + readOnlyGlobals.nfLitePluginNumPorts;

	if(pfring_add_filtering_rule(the_ring, &rule) < 0) {
	  traceEvent(TRACE_WARNING, "[NFLite] Unable to add PF_RING NFLite rule: quitting");
	  traceEvent(TRACE_WARNING, "[NFLite] Did you 'modprobe nflite_plugin' ?");
	  exit(-1);
	} else
	  traceEvent(TRACE_INFO, "[NFLite] PF_RING NFLite rule added successfully [UDP ports %d:%d]",
		     readOnlyGlobals.nfLitePluginLowPort,
		     readOnlyGlobals.nfLitePluginLowPort + readOnlyGlobals.nfLitePluginNumPorts);

	pfring_set_direction(the_ring, rx_only_direction);
	// pfring_set_cluster(the_ring, 1, cluster_per_flow);

	/* pfring_set_poll_watermark(the_ring, 0); */
      }

      pfring_enable_ring(the_ring);
    }
  }

  return(the_ring);
}

#endif /* HAVE_PF_RING */
